@@ -1,6 +1,6 @@
 /* client.c
  *
- * $Id: client.c,v 1.67 1999/06/14 01:23:49 nisse Exp $ */
+ * $Id: client.c,v 1.79 1999/10/11 20:53:31 nisse Exp $ */
 
 /* lsh, an implementation of the ssh protocol
  *
@@ -61,12 +61,15 @@
      (super packet_handler)
      (vars
        (service simple int)
-       (c object command_continuation)))
+       (c object command_continuation)
+       ;; Do we really need the exception handler here?
+       (e object exception_handler)))
 */
 
-static int do_accept_service(struct packet_handler *c,
-			     struct ssh_connection *connection,
-			     struct lsh_string *packet)
+static void
+do_accept_service(struct packet_handler *c,
+		  struct ssh_connection *connection,
+		  struct lsh_string *packet)
 {
   CAST(accept_service_handler, closure, c);
 
@@ -76,37 +79,42 @@ static int do_accept_service(struct packet_handler *c,
 
   simple_buffer_init(&buffer, packet->length, packet->data);
   
-  if ((parse_uint8(&buffer, &msg_number)
-      && (msg_number == SSH_MSG_SERVICE_ACCEPT))
-#if DATAFELLOWS_WORKAROUNDS
-      && (((connection->peer_flags & PEER_SERVICE_ACCEPT_KLUDGE) &&  parse_eod(&buffer)) ||
-#else
+  if (parse_uint8(&buffer, &msg_number)
+      && (msg_number == SSH_MSG_SERVICE_ACCEPT)
       && (
+#if DATAFELLOWS_WORKAROUNDS
+	  (connection->peer_flags & PEER_SERVICE_ACCEPT_KLUDGE)
+#else
+	  0
 #endif
-	  (parse_atom(&buffer, &name)
-	  && parse_eod(&buffer)
-	  && (name == closure->service))))
+	  || (parse_atom(&buffer, &name)
+	      && (name == closure->service)))
+      && parse_eod(&buffer))
     {
       lsh_string_free(packet);
       connection->dispatch[SSH_MSG_SERVICE_ACCEPT] = connection->fail;
       
-      return COMMAND_RETURN(closure->c, connection);
+      COMMAND_RETURN(closure->c, connection);
+    }
+  else
+    {
+      lsh_string_free(packet);
+      PROTOCOL_ERROR(closure->e, "Invalid SSH_MSG_SERVICE_ACCEPT message");
     }
-
-  lsh_string_free(packet);
-  return LSH_FAIL | LSH_DIE;
 }
 
 struct packet_handler *
 make_accept_service_handler(int service,
-			    struct command_continuation *c)
+			    struct command_continuation *c,
+			    struct exception_handler *e)
 {
   NEW(accept_service_handler, closure);
 
   closure->super.handler = do_accept_service;
   closure->service = service;
   closure->c = c;
-
+  closure->e = e;
+  
   return &closure->super;
 }
 
@@ -119,18 +127,20 @@ make_accept_service_handler(int service,
        ;; (service object ssh_service)))
 */
 
-static int do_request_service(struct command *s,
-			      struct lsh_object *x,
-			      struct command_continuation *c)
+static void
+do_request_service(struct command *s,
+		   struct lsh_object *x,
+		   struct command_continuation *c,
+		   struct exception_handler *e)
 {
   CAST(request_service, self, s);
   CAST(ssh_connection, connection, x);
   
   connection->dispatch[SSH_MSG_SERVICE_ACCEPT]
-    = make_accept_service_handler(self->service, c);
+    = make_accept_service_handler(self->service, c, e);
   
-  return A_WRITE(connection->write,
-		 format_service_request(self->service));
+  C_WRITE(connection,
+	  format_service_request(self->service));
 }
 
 struct command *make_request_service(int service)
@@ -184,7 +194,8 @@ struct command *make_request_service(int service)
 */
 
 /* Callback used when the server sends us eof */
-static int do_client_session_eof(struct ssh_channel *c)
+static void
+do_client_session_eof(struct ssh_channel *c)
 {
   CAST(client_session, session, c);
   
@@ -193,13 +204,15 @@ static int do_client_session_eof(struct ssh_channel *c)
   close_fd(&session->out->super, 0);
   close_fd(&session->err->super, 0);
 #endif
-
-  return LSH_OK;
 }  
 
-static int do_client_session_close(struct ssh_channel *c UNUSED)
+static void
+do_client_session_close(struct ssh_channel *c)
 {
-  return LSH_CHANNEL_PENDING_CLOSE;
+  static const struct exception finish_exception
+    = STATIC_EXCEPTION(EXC_FINISH_PENDING, "Session closed.");
+
+  EXCEPTION_RAISE(c->e, &finish_exception);
 }
 
 
@@ -211,11 +224,12 @@ static int do_client_session_close(struct ssh_channel *c UNUSED)
        (exit_status simple "int *")))
 */
 
-static int do_exit_status(struct channel_request *c,
-			  struct ssh_channel *channel,
-			  struct ssh_connection *connection UNUSED,
-			  int want_reply,
-			  struct simple_buffer *args)
+static void
+do_exit_status(struct channel_request *c,
+	       struct ssh_channel *channel,
+	       struct ssh_connection *connection UNUSED,
+	       int want_reply,
+	       struct simple_buffer *args)
 {
   CAST(exit_handler, closure, c);
   int status;
@@ -229,25 +243,23 @@ static int do_exit_status(struct channel_request *c,
       ALIST_SET(channel->request_types, ATOM_EXIT_STATUS, NULL);
       ALIST_SET(channel->request_types, ATOM_EXIT_SIGNAL, NULL);
 
-      /* Sent EOF, if we haven't done that already. */
+      /* Send EOF, if we haven't done that already. */
       /* FIXME: Make this behaviour configurable, there may be some
        * child process alive that we could talk to. */
 
-      if (!(channel->flags & CHANNEL_SENT_EOF))
-	return channel_eof(channel);
-      
-      return LSH_OK | LSH_GOON;
+      channel_eof(channel);
     }
-  
-  /* Invalid request */
-  return LSH_FAIL | LSH_DIE;
+  else
+    /* Invalid request */
+    PROTOCOL_ERROR(channel->e, "Invalid exit-status message");
 }
 
-static int do_exit_signal(struct channel_request *c,
-			  struct ssh_channel *channel,
-			  struct ssh_connection *connection UNUSED,
-			  int want_reply,
-			  struct simple_buffer *args)
+static void
+do_exit_signal(struct channel_request *c,
+	       struct ssh_channel *channel,
+	       struct ssh_connection *connection UNUSED,
+	       int want_reply,
+	       struct simple_buffer *args)
 {
   CAST(exit_handler, closure, c);
 
@@ -286,14 +298,11 @@ static int do_exit_signal(struct channel_request *c,
       /* FIXME: Make this behaviour configurable, there may be some
        * child process alive that we could talk to. */
 
-      if (!(channel->flags & CHANNEL_SENT_EOF))
-	return channel_eof(channel);
-
-      return LSH_OK | LSH_GOON;
+      channel_eof(channel);
     }
-  
-  /* Invalid request */
-  return LSH_FAIL | LSH_DIE;
+  else
+    /* Invalid request */
+    PROTOCOL_ERROR(channel->e, "Invalid exit-signal message");
 }
 
 struct channel_request *make_handle_exit_status(int *exit_status)
@@ -319,68 +328,107 @@ struct channel_request *make_handle_exit_signal(int *exit_status)
 }
 
 /* Receive channel data */
-static int do_receive(struct ssh_channel *c,
-		      int type, struct lsh_string *data)
+static void
+do_receive(struct ssh_channel *c,
+	   int type, struct lsh_string *data)
 {
   CAST(client_session, closure, c);
   
   switch(type)
     {
     case CHANNEL_DATA:
-      return A_WRITE(&closure->out->buffer->super, data);
+      A_WRITE(&closure->out->write_buffer->super, data);
+      break;
     case CHANNEL_STDERR_DATA:
-      return A_WRITE(&closure->err->buffer->super, data);
+      A_WRITE(&closure->err->write_buffer->super, data);
+      break;
     default:
       fatal("Internal error!\n");
     }
 }
 
 /* We may send more data */
-static int do_send(struct ssh_channel *c)
+static void
+do_send(struct ssh_channel *s,
+	struct ssh_connection *c UNUSED)
 {
-  CAST(client_session, closure, c);
+  CAST(client_session, self, s);
 
-  assert(closure->in->super.read);
-  assert(closure->in->handler);
-  closure->in->super.want_read = 1;
+  assert(self->in->super.read);
 
-  return LSH_OK | LSH_GOON;
+  self->in->super.want_read = 1;
 }
 
+#if 0
+/* Used for the first callback, to register stdin as a resource. */
+static void
+do_send_first(struct ssh_channel *s,
+	      struct ssh_connection *connection)
+{
+  CAST(client_session, self, s);
+
+  REMEMBER_RESOURCE(channel->resources, &self->in->super.super);
+  s->send = do_send;
+
+  do_send(s, connection);
+}
+#endif
+
 /* We have a remote shell */
-static int do_client_io(struct command *s UNUSED,
-			struct lsh_object *x,
-			struct command_continuation *c)
+static void
+do_client_io(struct command *s UNUSED,
+	     struct lsh_object *x,
+	     struct command_continuation *c,
+	     struct exception_handler *e UNUSED)
 
 {
-  if (!x)
-    {
-      werror("do_client_io: Starting shell failed.\n");
-      return LSH_CHANNEL_CLOSE | LSH_FAIL;
-    }
-  else
-    {
-      CAST(client_session, session, x);
-      struct ssh_channel *channel = &session->super;
+  CAST(client_session, session, x);
+  struct ssh_channel *channel = &session->super;
+  assert(x);
 
-      channel->receive = do_receive;
+  /* Set up write fd:s. */
   
-      session->out->super.close_callback
-	= session->err->super.close_callback = make_channel_close(channel);
+  channel->receive = do_receive;
+
+  /* FIXME: It seems a little kludgy to modify
+   * exception handlers here; it would be better to create the
+   * fd-objects at a point where the right exception handlers can be
+   * installed from the start. */
+  session->out->super.e
+    = make_channel_io_exception_handler(channel,
+					"lsh: I/O error on stdout",
+					session->out->super.e,
+					HANDLER_CONTEXT);
+
+  session->err->super.e
+    = make_channel_io_exception_handler(channel,
+					"lsh: I/O error on stderr",
+					session->err->super.e,
+					HANDLER_CONTEXT);
+
+  /* Set up the fd we read from. */
+  channel->send = do_send;
+
+  session->in->super.read = make_channel_read_data(channel);
+  if (session->super.send_window_size)
+    session->in->super.want_read = 1;
   
-      session->in->handler = make_channel_read_data(channel);
-      session->in->super.want_read = 1;
-      channel->send = do_send;
+  session->in->super.close_callback
+    = make_channel_read_close_callback(channel);
 
-      ALIST_SET(channel->request_types, ATOM_EXIT_STATUS,
-		make_handle_exit_status(session->exit_status));
-      ALIST_SET(channel->request_types, ATOM_EXIT_SIGNAL,
-		make_handle_exit_signal(session->exit_status));
+  /* Make sure stdio is closed properly if the channel or connection dies */
+  REMEMBER_RESOURCE(channel->resources, &session->in->super.super);
+  REMEMBER_RESOURCE(channel->resources, &session->out->super.super);
+  REMEMBER_RESOURCE(channel->resources, &session->err->super.super);
+  
+  ALIST_SET(channel->request_types, ATOM_EXIT_STATUS,
+	    make_handle_exit_status(session->exit_status));
+  ALIST_SET(channel->request_types, ATOM_EXIT_SIGNAL,
+	    make_handle_exit_signal(session->exit_status));
 
-      channel->eof = do_client_session_eof;
+  channel->eof = do_client_session_eof;
       
-      return COMMAND_RETURN(c, channel);
-    }
+  COMMAND_RETURN(c, channel);
 }
 
 struct command client_io =
@@ -415,8 +463,8 @@ struct ssh_channel *make_client_session(struct io_fd *in,
   self->err = err;
 
   /* Flow control */
-  out->buffer->report = &self->super.super;
-  err->buffer->report = &self->super.super;
+  out->write_buffer->report = &self->super.super;
+  err->write_buffer->report = &self->super.super;
   
   self->exit_status = exit_status;
   
@@ -443,7 +491,7 @@ new_session(struct channel_open_command *s,
 
   self->session->write = connection->write;
   
-  *request = prepare_channel_open(connection->channels, ATOM_SESSION,
+  *request = prepare_channel_open(connection, ATOM_SESSION,
 				  self->session, "");
   if (!*request)
     return NULL;
