@@ -1,6 +1,6 @@
 /* connection_commands.c
  *
- * $Id: connection_commands.c,v 1.6 1999/06/22 22:48:23 nisse Exp $
+ * $Id: connection_commands.c,v 1.15 1999/10/24 20:12:31 nisse Exp $
  */
 
 /* lsh, an implementation of the ssh protocol
@@ -50,7 +50,8 @@
        (connection object ssh_connection)))
 */
 
-static int connection_die(struct close_callback *c, int reason)
+static void
+connection_die(struct close_callback *c, int reason)
 {
   CAST(connection_close_handler, closure, c);
   
@@ -59,8 +60,6 @@ static int connection_die(struct close_callback *c, int reason)
     werror("Connection died.\n");
 
   KILL_RESOURCE_LIST(closure->connection->resources);
-  
-  return 4711;  /* Ignored */
 }
 
 struct close_callback *make_connection_close_handler(struct ssh_connection *c)
@@ -85,10 +84,14 @@ struct close_callback *make_connection_close_handler(struct ssh_connection *c)
        (fallback object ssh1_fallback)))
 */
 
-static int split_version_string(UINT32 length, UINT8 *line,
-				UINT32 *protover_len, UINT8 **protover,
-				UINT32 *swver_len, UINT8 **swver,
-				UINT32 *comment_len, UINT8 **comment)
+/* Returns -1 if the line is not the start of a SSH handshake,
+ * 0 if the line appears to be an SSH handshake, but with bogus version fields,
+ * or 1 if the line was parsed sucessfully. */
+static int
+split_version_string(UINT32 length, UINT8 *line,
+		     UINT32 *protover_len, UINT8 **protover,
+		     UINT32 *swver_len, UINT8 **swver,
+		     UINT32 *comment_len, UINT8 **comment)
 {
   UINT8 *sep;
 
@@ -109,6 +112,8 @@ static int split_version_string(UINT32 length, UINT8 *line,
   line = sep + 1;
   length -= *protover_len + 1;
 
+  /* FIXME: The spec is not clear about the separator here. Can there
+   * be other white space than a single space character? */
   sep = memchr(line, ' ', length);
   if (!sep)
     {
@@ -126,15 +131,17 @@ static int split_version_string(UINT32 length, UINT8 *line,
   return 1;
 }
 
-static int do_line(struct line_handler **h,
-		   struct read_handler **r,
-		   UINT32 length,
-		   UINT8 *line)
+static void
+do_line(struct line_handler **h,
+	struct read_handler **r,
+	UINT32 length,
+	UINT8 *line,
+	struct exception_handler *e)
 {
   CAST(connection_line_handler, closure, *h);
   UINT32 protover_len, swver_len, comment_len;
   UINT8 *protover, *swver, *comment;
-
+  
   switch(split_version_string(length, line, 
 			      &protover_len, &protover, 
 			      &swver_len, &swver,
@@ -147,6 +154,7 @@ static int do_line(struct line_handler **h,
 	 * to accept a client that wants version 1.99. But Datafellow's
 	 * ssh2 client does exactly that, so we have to support it. And
 	 * I don't think it causes any harm. */
+	
 	if ( ((protover_len >= 3) && !memcmp(protover, "2.0", 3))
 	     || ((protover_len == 4) && !memcmp(protover, "1.99", 4)) )
 	  {
@@ -154,43 +162,38 @@ static int do_line(struct line_handler **h,
 #if WITH_SSH1_FALLBACK
 	    if (closure->fallback)
 	      {
-		int res;
-		
 		assert(closure->mode == CONNECTION_SERVER);
-		
+	      
 		/* Sending keyexchange packet was delayed. Do it now */
-		res = initiate_keyexchange(closure->connection,
-					   closure->mode);
-		
-		if (LSH_CLOSEDP(res))
-		  {
-		    werror("server.c: do_line: "
-			   "Delayed initiate_keyexchange() failed.\n");
-		    *h = NULL;
-		    
-		    return res | LSH_DIE;
-		  }
+		initiate_keyexchange(closure->connection,
+				     closure->mode);
 	      }
 #endif /* WITH_SSH1_FALLBACK */
 
 #if DATAFELLOWS_WORKAROUNDS
-	    if (swver_len > 4 && memcmp(swver, "2.0.", 4) == 0 &&
-		memcmp(swver + 4, "13", 2) <= 0)
+	    if ( (swver_len >= 6) && 
+		 (memcmp(swver, "2.0.", 4) == 0) &&
+		 /* FIXME: Perhaps do a numerical comparison here? */
+		 (memcmp(swver + 4, "13", 2) <= 0) )
 	      {
-		closure->connection->peer_flags = PEER_SSH_DSS_KLUDGE | PEER_SERVICE_ACCEPT_KLUDGE | PEER_USERAUTH_REQUEST_KLUDGE;
+		closure->connection->peer_flags
+		  |= (PEER_SSH_DSS_KLUDGE | PEER_SERVICE_ACCEPT_KLUDGE
+		      | PEER_USERAUTH_REQUEST_KLUDGE | PEER_SEND_NO_DEBUG);
 	      }
 #endif	    
+	    
 	    new = 
 	      make_read_packet(
-	        make_packet_unpad(
-	          make_packet_inflate(
-	            make_packet_debug(&closure->connection->super, "received"),
-	            closure->connection
-	          )
-	        ),
-	        closure->connection
-	      );
-
+		make_packet_unpad(
+		  closure->connection,
+		  make_packet_inflate(
+		    make_packet_debug(&closure->connection->super, "received"),
+		    closure->connection
+		    )
+		  ),
+		closure->connection
+		);
+	    
 	    closure->connection->versions[!closure->mode]
 	      = ssh_format("%ls", length, line);
 
@@ -200,51 +203,56 @@ static int do_line(struct line_handler **h,
 		    closure->connection->versions[CONNECTION_SERVER]);
 
 	    *r = new;
-	    return LSH_OK | LSH_GOON;
+	    return;
 	  }
 #if WITH_SSH1_FALLBACK      
-	else if (closure->fallback
-		 && (protover_len >= 2)
-		 && !memcmp(protover, "1.", 2))
-	  {
-	    *h = NULL;
-	    return SSH1_FALLBACK(closure->fallback,
-				 closure->fd,
-				 length, line);
-	  }
-#endif /* WITH_SSH1_FALLBACK */
-	else
-	  {
-	    werror("Unsupported protocol version: %ps\n",
-		   length, line);
+      else if (closure->fallback
+	       && (protover_len >= 2)
+	       && !memcmp(protover, "1.", 2))
+	{
+	  *h = NULL;
+	  SSH1_FALLBACK(closure->fallback,
+			closure->fd,
+			length, line,
+			e);
 
-	    /* FIXME: Clean up properly */
-	    KILL(closure);
-	    *h = NULL;
-	  
-	    return LSH_FAIL | LSH_DIE;
-	  }
+	  return;
+	}
+#endif /* WITH_SSH1_FALLBACK */
+      else
+	{
+	  werror("Unsupported protocol version: %ps\n",
+		 length, line);
+
+	  /* FIXME: Clean up properly */
+	  KILL(closure);
+	  *h = NULL;
+
+	  EXCEPTION_RAISE(closure->connection->e,
+			  make_protocol_exception
+			  (SSH_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED,
+			   NULL));
+	  return;
+	}
+	fatal("Internal error!\n");
       }
     case 0:
-      {
-	werror("Incorrectly formatted version string: %s\n", length, line);
-	KILL(closure);
-	*h = NULL;
-	return LSH_FAIL | LSH_DIE;
-      }
+      werror("Incorrectly formatted version string: %s\n", length, line);
+      KILL(closure);
+      *h = NULL;
+      
+      PROTOCOL_ERROR(closure->connection->e,
+		     "Incorrectly version string.");
+      
+      return;
     case -1:
-      {
-	/* Display line */
-	werror("%ps\n", length, line);
-	
-	/* Read next line */
-	return LSH_OK | LSH_GOON;
-      }
+      /* Display line */
+      werror("%ps\n", length, line);
+
+      /* Read next line */
+      break;
     default:
-      {
-	werror("Internal error!\n");
-	return LSH_FAIL | LSH_DIE;
-      }
+      fatal("Internal error!\n");
     }
 }
 
@@ -260,7 +268,7 @@ make_connection_read_line(struct ssh_connection *connection, int mode,
   closure->mode = mode;
   closure->fd = fd;
   closure->fallback = fallback;
-  return make_read_line(&closure->super);
+  return make_read_line(&closure->super, connection->e);
 }
 
 /* Takes a fd as argument, and returns a connection object. Never
@@ -284,17 +292,20 @@ make_connection_read_line(struct ssh_connection *connection, int mode,
        (fallback object ssh1_fallback)))
 */
 
-static int do_connection(struct command *s,
-			 struct lsh_object *x,
-			 struct command_continuation *c)
+/* Buffer size when reading from the socket */
+#define BUF_SIZE (1<<14)
+
+static void
+do_connection(struct command *s,
+	      struct lsh_object *x,
+	      struct command_continuation *c,
+	      struct exception_handler *e)
 {
   CAST(connection_command, self, s);
   CAST(io_fd, fd, x);
   struct lsh_string *version;
-  int res;
-  
-  struct ssh_connection *connection = make_ssh_connection(c);
-
+  struct ssh_connection *connection;
+    
   switch (self->mode)
     {
     case CONNECTION_CLIENT:
@@ -325,14 +336,29 @@ static int do_connection(struct command *s,
       fatal("do_connection: Internal error\n");
     }
 
+  /* Installing the right exception handler is a little tricky. The
+   * passed in handler is typically the top-level handler provided by
+   * lsh.c or lshd.c. On top of this, we add the io_exception_handler
+   * which takes care of EXC_FINISH_READ exceptions and closes the
+   * connection's socket. And on top of this, we have a
+   * connection_exception handler, which takes care of EXC_PROTOCOL
+   * exceptions, sends a disconnect message, and then raises an
+   * EXC_FINISH_READ exception. */
+  
+  connection = make_ssh_connection
+    (c, make_exc_finish_read_handler(&fd->super,
+				     e, HANDLER_CONTEXT));
+  
   connection_init_io
     (connection, 
      &io_read_write(fd,
-		    make_connection_read_line(connection, self->mode,
-					      fd->super.fd, self->fallback),
+		    make_buffered_read
+		    (BUF_SIZE,
+		     make_connection_read_line(connection, self->mode,
+					       fd->super.fd, self->fallback)),
 		    self->block_size,
 		    make_connection_close_handler(connection))
-     ->buffer->super,
+     ->write_buffer->super,
      self->random);
 
   connection->versions[self->mode] = version;
@@ -347,16 +373,17 @@ static int do_connection(struct command *s,
    * Furthermore, it should not send any data after the identification string,
    * until the client's identification string is received. */
   if (self->fallback)
-    return A_WRITE(connection->raw,
-		   ssh_format("%lS\n", version));
+    {
+      A_WRITE(connection->raw,
+	      ssh_format("%lS\n", version));
+      return;
+    }
 #endif /* WITH_SSH1_FALLBACK */
 
-  res = A_WRITE(connection->raw,
-		ssh_format("%lS\r\n", version));
-  if (LSH_CLOSEDP(res))
-    return res;
+  A_WRITE(connection->raw,
+	  ssh_format("%lS\r\n", version));
   
-  return res | initiate_keyexchange(connection, self->mode);
+  initiate_keyexchange(connection, self->mode);
 }
 
 struct command *
