@@ -2,7 +2,7 @@
  *
  *
  *
- * $Id: publickey_crypto.h,v 1.21 1999/06/14 00:37:49 nisse Exp $ */
+ * $Id: publickey_crypto.h,v 1.24 1999/09/22 01:11:39 nisse Exp $ */
 
 /* lsh, an implementation of the ssh protocol
  *
@@ -35,15 +35,31 @@
 #include "publickey_crypto.h.x"
 #undef GABA_DECLARE
 
+/* GABA:
+   (class
+     (name keypair)
+     (vars
+       ; atom identifying algorithm type
+       (type . UINT32)
+       (public string)
+       (private object signer)))
+*/
+
+struct keypair *make_keypair(UINT32 type,
+			     struct lsh_string *public,
+			     struct signer *private);
+
 /* parse an ssh keyblob */
 int parse_dsa_public(struct simple_buffer *buffer,
 		     struct dsa_public *public);
+void init_dsa_public(struct dsa_public *public);
 
 struct signature_algorithm *make_dsa_algorithm(struct randomness *random);
+
 #if DATAFELLOWS_WORKAROUNDS
 struct signer *make_dsa_signer_kludge(struct signer *dsa);
 struct verifier *make_dsa_verifier_kludge(struct verifier *v);
-struct signature_algorithm *make_dsa_kludge_algorithm(struct randomness *random);
+/* struct signature_algorithm *make_dsa_kludge_algorithm(struct randomness *random); */
 #endif
 
 struct signer *make_dsa_signer_classic(struct signer *s);
