@@ -22,7 +22,8 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include "server_publickey.h"
+#include "server_userauth.h"
+#include "userauth.h"
 #include "xalloc.h"
 #include "connection.h"
 #include "parse.h"
@@ -31,7 +32,7 @@
 #include "charset.h"
 #include "ssh.h"
 #include "lookup_verifier.h"
-#include "server_password.h"
+#include "publickey_crypto.h"
 
 #include "server_publickey.c.x"
 
@@ -49,53 +50,67 @@ static struct lsh_string *format_userauth_pk_ok(int algorithm, struct lsh_string
 		    algorithm, keyblob);
 }
 
-static int do_authenticate(struct userauth *i UNUSED,
-			   struct ssh_connection *connection,
-			   struct lsh_string *username,
-			   struct simple_buffer *args,
-			   struct command_continuation *c)
+static void
+do_authenticate(struct userauth *s,
+		struct ssh_connection *connection,
+		struct lsh_string *username,
+		struct simple_buffer *args,
+		struct command_continuation *c,
+		struct exception_handler *e)
 {
-  CAST(userauth_publickey, self, i);
+  CAST(userauth_publickey, self, s);
+  
   struct lsh_string *keyblob = NULL;
   struct lookup_verifier *lookup;
   struct verifier *v;
   UINT8 *signature_blob;
   UINT32 signature_length;
-  UINT32 signature_start;
+  UINT32 signature_start = 0;
   UINT32 algorithm;
   int check_key;
 
   username = utf8_to_local(username, 1, 1);
   if (!username)
-    return LSH_FAIL | LSH_DIE;
+    {
+      PROTOCOL_ERROR(e, "Invalid utf8 in username.");
+      return;
+    }
 
   if (parse_boolean(args, &check_key) &&
       parse_atom(args, &algorithm) &&
       (keyblob = parse_string_copy(args)) &&
-      ((check_key && 
-	(signature_start = args->pos) && 
-	(parse_string(args, &signature_length, &signature_blob)) &&
-	parse_eod(args)) || 
-       (!check_key && parse_eod(args)))) 
+      (check_key
+       ? ((signature_start = args->pos)
+	  , (parse_string(args, &signature_length, &signature_blob))
+	  && parse_eod(args))
+       : parse_eod(args))) 
     {
-#if DATAFELLOWS_WORKAROUNDS
-      if (algorithm == ATOM_SSH_DSS && (connection->peer_flags & PEER_SSH_DSS_KLUDGE))
-	{
-	  lookup = ALIST_GET(self->verifiers, ATOM_SSH_DSS_KLUDGE);
-	}
-      else
-#endif
-	lookup = ALIST_GET(self->verifiers, algorithm);
+
+      lookup = ALIST_GET(self->verifiers, algorithm);
 
       if (!lookup) 
 	{
+	  static const struct exception unsupported_publickey_algorithm
+	    = STATIC_EXCEPTION(EXC_USERAUTH,
+			       "Unsupported public key algorithm.");
+	  
 	  lsh_string_free(keyblob); 
 	  lsh_string_free(username);
 	  verbose("Unknown publickey algorithm\n");
-	  return COMMAND_RETURN(c, NULL);
+	  EXCEPTION_RAISE(e, &unsupported_publickey_algorithm);
+	  return;
 	}
       v = LOOKUP_VERIFIER(lookup, username, keyblob);
 
+#if DATAFELLOWS_WORKAROUNDS
+      if (v && 
+	  (algorithm == ATOM_SSH_DSS) && 
+	  (connection->peer_flags & PEER_SSH_DSS_KLUDGE))
+	{
+	  v = make_dsa_verifier_kludge(v);
+	}
+#endif
+
       if (!check_key)
 	{
 	  lsh_string_free(username);
@@ -104,20 +119,24 @@ static int do_authenticate(struct userauth *i UNUSED,
 	      struct lsh_string *reply = format_userauth_pk_ok(algorithm, keyblob);
 	      lsh_string_free(keyblob);
 	      KILL(v);
-	      return A_WRITE(connection->write, reply);
+	      EXCEPTION_RAISE(e, make_userauth_special_exception(reply, NULL));
+	      return;
 	    }
 	  else
 	    {
+	      static const struct exception unauthorized_key
+		= STATIC_EXCEPTION(EXC_USERAUTH,
+				   "Unauthorized public key.");
+	      
 	      lsh_string_free(keyblob);
-	      return COMMAND_RETURN(c, NULL);
+	      EXCEPTION_RAISE(e, &unauthorized_key);
+	      return;
 	    }
 	}
       else 
 	{
 	  struct lsh_string *signed_data;
-	  struct unix_user *user = NULL;
 #if DATAFELLOWS_WORKAROUNDS
-	  /* FIXME: this check should be made more general */
 	  if (connection->peer_flags & PEER_USERAUTH_REQUEST_KLUDGE)
 	    {
 	      signed_data = ssh_format("%lS%c%S%a%a%c%a%S", 
@@ -132,24 +151,35 @@ static int do_authenticate(struct userauth *i UNUSED,
 	    }
 	  else
 #endif
-	      signed_data = ssh_format("%lS%ls", connection->session_id, 
-				       signature_start, args->data);
+	    /* The signature is on the session id, followed by the
+	     * userauth request up to the actual signature. To avoid collisions,
+	     * the length field for the session id is included. */
+	    signed_data = ssh_format("%S%ls", connection->session_id, 
+				     signature_start, args->data);
 
 	  lsh_string_free(keyblob); 
 	  if (VERIFY(v, signed_data->length, signed_data->data, signature_length, signature_blob))
 	    {
 	      werror("user %S has authenticated\n", username);
-	      user = lookup_user(username, 0);
+	      COMMAND_RETURN(c, lookup_user(username, 0));
+	    }
+	  else
+	    {
+              static const struct exception bad_sign
+                = STATIC_EXCEPTION(EXC_USERAUTH, "Bad signature in userauth request");
+	      
+              EXCEPTION_RAISE(e, &bad_sign);
+              return;
 	    }
 	  lsh_string_free(username);
 	  lsh_string_free(signed_data);
-	  return COMMAND_RETURN(c, user);
+	  return;
 	}
     }
   else 
     {
       werror("Badly formatted publickey authentication request\n");
-      return LSH_FAIL | LSH_DIE;
+      PROTOCOL_ERROR(e, "Invalid publickey authentication request");
     }
 }
 
